
#include "OggVorbis.h"


// #define DEBUG

// if version is not defined - define it manually
#ifndef VORBIS_SDK
 #define VORBIS_SDK "/vorbis-sdk-1.0rc2/"
#endif



// buffer size
#define PCMCHUNKSIZE	32768
#define APOLLO_REGISTRY	"Software\\Capacala\\Apollo"
#define APOLLO_REGYSTRY_TEMPL	"trackTitleTemplate"

#define MAX_NAME 256

//#define SAMPLERATE	  44100
//#define BITRESOLUTION   16
//#define NUMCHANNELS	 2

#ifdef DEBUG
 #define DBG(txt,cap) MessageBox(GetActiveWindow(),txt,cap,MB_OK)
#else
 #define DBG(txt,cap)
#endif


////////////////////////////
//
//	what makes us plugin
//

extern "C" __declspec(dllexport) CInputPlugin *Apollo_GetInputModule2(HINSTANCE hDLLInstance,HWND hApolloWnd)
{
	// DBG("Call Apollo_GetInputModule2", "DBG");
	return new COggVorbisPlugin(hDLLInstance,hApolloWnd);
}



///////////////////////////
//
//	Vorbis Callbacks
//

size_t vfcb_i_read (void *ptr, size_t size, size_t nmemb, void *datasource)
{
	HANDLE * _hFile = (HANDLE *) datasource;

	unsigned long read = 0;
	if (ReadFile(*_hFile, ptr, size*nmemb, &read, NULL)) {
		// number of read==0 indicated no more data
		return read;
	} else {
		return 0;
	};

	return FALSE;
};

int vfcb_i_seek(void *datasource, ogg_int64_t offset, int whence)
{
	HANDLE * _hFile = (HANDLE *) datasource;
	LARGE_INTEGER li;
	li.QuadPart = offset;
	return !! SetFilePointer(*_hFile, li.LowPart, &li.HighPart, whence);
};

long vfcb_i_tell(void *datasource)
{
	HANDLE * _hFile = (HANDLE *) datasource;

	LARGE_INTEGER li;
	li.QuadPart = 0;

	SetLastError(NO_ERROR);
	li.LowPart = SetFilePointer(*_hFile, li.LowPart, &li.HighPart, FILE_CURRENT);
	if (GetLastError() == NO_ERROR)
		return li.QuadPart;	// success

	return 0;
};

int vfcb_i_close(void *datasource)
{
	HANDLE * _hFile = (HANDLE *) datasource;
	CloseHandle(*_hFile);
        *_hFile = INVALID_HANDLE_VALUE;
	return TRUE;
};




void get_suggested_title(OggVorbis_File * vf, char * fn, char * outbuf, size_t buf_len)
{
	// "%LEADARTIST%  %TITLE% // %COMMENT%"


/*	Dump from Apollo.exe
	hate proprietary/closed licenses/sources
	!Vive GPL!
	!Anarchy!-)))


F F F ȊF ԊF NOFRAME ORIGALBUM   PUBLISHER   ENCODEDBY   ENCODERSETTINGS ORIGFILENAME    LANGUAGE
PARTINSET   DATE    TIME    RECORDINGDATES  MEDIATYPE   FILETYPE    NETRADIOSTATION NETRADIOOWNER
LYRICIST    ORIGARTIST  ORIGLYRICIST    SUBTITLE    MIXARTIST   USERTEXT    CONTENTGROUP    TITLE
LEADARTIST  BAND    ALBUM   YEAR    CONDUCTOR   COMPOSER    COPYRIGHT   CONTENTTYPE TRACKNUM    COMMEN
T WWWAUDIOFILE    WWWARTIST   WWWAUDIOSOURCE  WWWCOMMERCIALINFO   WWWCOPYRIGHT    WWWPUBLISHER    WWWP
AYMENT  WWWRADIOPAGE    WWWUSER INVOLVEDPEOPLE  UNSYNCEDLYRICS  PICTURE GENERALOBJECT   UNIQUEFILEID
  PLAYCOUNTER POPULARIMETER   GROUPINGREG CRYPTOREG   No frame    Original album/movie/show title Publ
isher   Encoded by  Software/Hardware and settings used for encoding    Original filename   Language(s
) Part of a set   Date    Time    Recording dates Media type  File type   Internet radio station name
Internet radio station owner    Lyricist/Text writer    Original artist(s)/performer(s) Original lyric
ist(s)/text writer(s) Subtitle/Description refinement Interpreted, remixed, or otherwise modified by
User defined text information frame Content group description   Title/songname/content description  Le
ad performer(s)/Soloist(s)    Band/orchestra/accompaniment    Album/Movie/Show title  Year    Conducto
r/performer refinement  Composer    Copyright message   Content type    Track number/Position in set
  Comments    Official audio file webpage Official artist/performer webpage   Official audio source we
bpage   Commercial information  Copyright/Legal information Publishers official webpage Payment Offici
al internet radio station homepage    User defined URL link frame Involved people list    Unsychronize
d lyric/text transcription  Attached picture    General encapsulated object Unique file identifier  Pl
ay counter    Popularimeter   Group identification registration   Encryption method registration  TAG
eng ID3v1 comment field %d  (%d)    ID3Lib v3.05a said "%s", but we couldn't care less. Apollo  Blues
*/

/* sample vorbis comments:

artist=The USSR Symphony Orch. cond Evgeni Svetlanov
date=1968
album=S. Rachmaninov, Symphony No. 3
tracknumber=05
title=Scherzo for Orchestra
genre=Classics

*/
	char templ[MAX_NAME];

	DBG("come into get_suggested_title","DBG");

	HKEY rk;
	if (RegOpenKeyEx(HKEY_CURRENT_USER, APOLLO_REGISTRY, 0, KEY_READ, &rk) != ERROR_SUCCESS) {
		DBG("Failed to open reg", "HKEY_CURRENT_USER\\" APOLLO_REGISTRY);
		strcpy(outbuf, fn);
		return;
	}

	DWORD type;
//	unsigned char data[MAX_NAME];
	DWORD data_length = sizeof(templ);
	bool suc = RegQueryValueEx(rk, APOLLO_REGYSTRY_TEMPL, NULL, &type, (unsigned char *) &templ, &data_length) == ERROR_SUCCESS;

	RegCloseKey(rk);

	if (!suc) {
		DBG("Failed to read reg", "HKEY_CURRENT_USER\\" APOLLO_REGISTRY APOLLO_REGYSTRY_TEMPL);
		strcpy(outbuf, fn);
		return;
	};

	DBG("before vf2id","DBG");

	char * vf2id[][2] = {	// to translate Vorbis tags namespace into ID3 namespace
		{"", "comment"},
		{"comments", "comment"},
		{"artist", "leadartist"},
		{"tracknumber", "tracknum"},
		{NULL, NULL},	// stop!
	};

	char ** ptr=ov_comment(vf,-1)->user_comments;

	//
	// go thru list of Vorbis tags
	// split // translate to ID3 names
	//

	int tags_num = 0;
	{	// count tags to allocate memory
		char ** p = ptr;
		while (*p) { tags_num++; p++; };
	};

	// storage for tags names/values - it allows us fix names
	char (* tags) [][2][MAX_NAME] = (char (*) [][2][MAX_NAME]) malloc(tags_num*2*MAX_NAME);

	memset(*tags, 0, tags_num*2*MAX_NAME);

	int i=0;
	DBG("before copy 2 storage","DBG");
	while(*ptr) {	// loop^2, any better idea?

		char *param_name = (*tags)[i][0];
		char *param_val  = (*tags)[i][1];

		// split by '=';
		if (char *delim = strchr(*ptr, '=')) {
			// copy to storage
			strncpy(param_name, *ptr, min(MAX_NAME, delim - *ptr));
			strncpy(param_val, delim+1, MAX_NAME);

			char * (* foo)[2] = &vf2id[0];	// not too strong in C - hope this is correct...

			// apply name fix if needed
			while (*foo[0]) {
				if (strnicmp(param_name, (*foo)[0], MAX_NAME) == 0) {	// if eq
					strncpy(param_name, (*foo)[1], MAX_NAME);	// assign new name to param
					break;	// first entry
				};
				foo++;
			};
		};

		i++; ptr++; // go on
	};

	DBG("before extract %","DBG");
	//
	// extract %.+% from templ
	//
	char (* foo) = templ;
	char printf_ctrl[MAX_NAME];
	memset(printf_ctrl,0,sizeof(printf_ctrl));

	char (* bar) = printf_ctrl;
	char (* needed_tags) = (char (*)) malloc(strlen(templ)); // safe - ctrl str could be only smaller
	memset(needed_tags, 0, strlen(templ));
	char (* curr_nt) = needed_tags;

	int nt_num = 0;	// needed tags num

	while (*foo != '\00') {
		if (*foo == '%') {
			*bar = *foo; bar++;
			if (*(foo+1) == '%') {
				*bar = *foo;	// put %% in printf_ctrl
				bar++; foo++; // skip %%
			} else {
				if (char * baz = strchr(foo+1, '%')) { // matched %
					*bar = 's';	// make "...%s..."
					bar++;
					// copy tag name to array
					int jump = baz-foo-1;
					strncpy(curr_nt, foo+1, jump);
					DBG(curr_nt, "MATCH curr_nt");
					curr_nt += jump+1;	// leave one \00
					foo = baz;	// jump to matching '%'
					nt_num++;
				} else {
					*bar = '%';	// leave it matched
					bar++;
				};
			};
		} else {
			*bar = *foo;
			bar++;
		};

		foo++;
	};

	char *(* printf_params)[] = (char *(*)[]) malloc(sizeof(char *)*nt_num);

        curr_nt = needed_tags;

        //
        //  go thru tags, tags <=> needed_tags => printf_params
        //

        // for each entry in needed_tags find corr entry in tags.

        char * dummy = "";	// empty string for unknown entries

        int j=0;
	while (*curr_nt != '\00') {	// loop^2

		// (*printf_params)[j] = curr_nt;
		int i=0;
		for (;i<tags_num; i++) {
			DBG((*tags)[i][0], curr_nt);
			if ( strnicmp( (*tags)[i][0], curr_nt, MAX_NAME) == 0)
				break;
		};

                (*printf_params)[j] = (i==tags_num) ? dummy : (*tags)[i][1];

		j++;
		curr_nt += strlen(curr_nt)+1;	// len + '\00'
	};

	DBG(printf_ctrl, "printf_ctrl");


//	char localbuf[1025];	// Windoz wvsprintf has limit of 1k so it is sane

//	wvsprintf(localbuf, printf_ctrl, (char *) *printf_params);

        _vsnprintf(outbuf, buf_len, printf_ctrl, (char *) *printf_params);

	// trying to be safe on out buf length
//	strncpy(outbuf, localbuf, buf_len);

	free(printf_params);

	free(needed_tags);

	free(tags);

};






////////////////////////////
//
//	COggVorbisPlugin methods defenition
//




char *COggVorbisPlugin::GetDescription()
{
	return "OggVorbis Plug-in by Philips";
}

CInputDecoder *COggVorbisPlugin::Open(char *filename,int audioDataOffset)
{
	DBG("Call COggVorbisPlugin::Open", filename);
	return new COggVorbisDecoder(filename,audioDataOffset);
}

BOOL COggVorbisPlugin::GetInfo(char *filename,int audioDataOffset,TrackInfo *trackInfo)
{

	OggVorbis_File vf;
        vorbis_info * vi;

	ov_callbacks _cb;

	// local needs
       	_cb.read_func = &vfcb_i_read;
	_cb.seek_func = &vfcb_i_seek;
	_cb.close_func = &vfcb_i_close;
	_cb.tell_func = &vfcb_i_tell;


        if (!filename || !trackInfo) return FALSE;

	DBG("Call COggVorbisPlugin::GetInfo", filename);
//	return FALSE;
	if (strstr(filename,"://") != NULL) return FALSE;   // we can't handle URLs

	HANDLE hFile = CreateFile(filename,GENERIC_READ,FILE_SHARE_READ | FILE_SHARE_WRITE,NULL,OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL,NULL);

	if (hFile == INVALID_HANDLE_VALUE) {
       		MessageBox(GetActiveWindow(), "failed to CreateFile", filename, 0);
		return FALSE;
	};

	if (ov_open_callbacks((void *) &hFile, &vf, NULL, 0, _cb)) {	// get error
		MessageBox(GetActiveWindow(), "failed to ov_open", filename, 0);
		ov_clear(&vf);
		CloseHandle(hFile);
		return FALSE;
	} else {
		DBG("success to ov_open", filename);
	};


	vi = ov_info(&vf, -1);

	if (!vi) {
		MessageBox(GetActiveWindow(), "failed to ov_info", filename, 0);
		ov_clear(&vf);
		CloseHandle(hFile);
		return FALSE;
	} else {
		DBG("success to ov_info", filename);
	};

//	MessageBox(GetActiveWindow(), "Before main block", "DBG", 0);
	trackInfo->fileSize = GetFileSize(hFile,NULL);
	trackInfo->seekable = ov_seekable(&vf);
	trackInfo->hasEqualizer = FALSE;	// we want the player to provide equalization
	trackInfo->sampleRate = vi->rate;
	trackInfo->bitResolution = 16;	// ? where to get ?
	trackInfo->numChannels = vi->channels;
	trackInfo->playingTime = (int) (ov_time_total(&vf, -1)*1000);

	trackInfo->bitRate = vi->bitrate_nominal;
	if (trackInfo->bitRate == 0) {
	    if (trackInfo->playingTime != 0) {
		// Calculate an approximate
		trackInfo->bitRate = (int) ((__int64) trackInfo->fileSize * 8 * 1000 / trackInfo->playingTime);
	    }
	    if (trackInfo->bitRate == 0) {
		// Apollo can't handle zero bitrate => set to one
		trackInfo->bitRate = 1;
	    }
	}
//	MessageBox(GetActiveWindow(), "In mid of main block", "DBG", 0);

	strcpy(trackInfo->fileTypeDescription, "OggVorbis File");

	char name_buf[MAX_NAME];

        get_suggested_title(&vf, filename, name_buf, sizeof(name_buf));
//	strncpy(name_buf, filename, MAX_NAME);

	strncpy(trackInfo->suggestedTitle, name_buf, sizeof(trackInfo->suggestedTitle));
//	MessageBox(GetActiveWindow(), "After main block", "DBG", 0);

	ov_clear(&vf);
	CloseHandle(hFile);

	return TRUE;

/*
	HANDLE hFile = CreateFile(filename,GENERIC_READ,FILE_SHARE_READ | FILE_SHARE_WRITE,NULL,OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL,NULL);

	if (hFile == INVALID_HANDLE_VALUE) return FALSE;	// error opening file

	trackInfo->fileSize = GetFileSize(hFile,NULL);

	CloseHandle(hFile);

	trackInfo->seekable = TRUE;
	trackInfo->hasEqualizer = FALSE;		// we want the player to provide equalization
	trackInfo->sampleRate = SAMPLERATE;
	trackInfo->bitResolution = BITRESOLUTION;
	trackInfo->numChannels = NUMCHANNELS;
	trackInfo->playingTime = (int)((__int64)(trackInfo->fileSize - audioDataOffset) * 1000 / (SAMPLERATE * NUMCHANNELS * (BITRESOLUTION >> 3)));
	trackInfo->bitRate = SAMPLERATE * NUMCHANNELS * BITRESOLUTION;
		// if the bitrate is not constant, store the average value with a negative sign

	strcpy(trackInfo->fileTypeDescription,"Raw PCM File");

	// if the file contains information about the title you should
	// use it as the suggested title instead of the filebody

	char *extensionBegin = strrchr(filename,'.');

	if (extensionBegin == NULL)
		extensionBegin = filename + strlen(filename);

	char *pathEnd = strrchr(filename,'\\');

	if (pathEnd == NULL)
	{
		memcpy(trackInfo->suggestedTitle,filename,extensionBegin - filename);
		trackInfo->suggestedTitle[extensionBegin - filename] = 0;
	}
	else
	{
		memcpy(trackInfo->suggestedTitle,pathEnd + 1,extensionBegin - pathEnd - 1);
		trackInfo->suggestedTitle[extensionBegin - pathEnd - 1] = 0;
	}
*/

//	return TRUE;
}

void COggVorbisPlugin::AdditionalInfo(char *filename,int audioDataOffset)
{
	// opens a dialog/window containing extra information about the file

	// shame on me...
	MessageBox(GetActiveWindow(), "Not implemented","Additional Info...", 0);
}

BOOL COggVorbisPlugin::IsSuitableFile(char *filename,int audioDataOffset)
{
	// here you should detect whether this plug-in supports the file or not

	if (strstr(filename,"://") != NULL) return FALSE;   // we can handle URLs

	char *extensionBegin = strrchr(filename,'.');

	if (extensionBegin == NULL) return FALSE;

	if (!stricmp(extensionBegin,".ogg")) return TRUE;
//	if (!stricmp(extensionBegin,".pcm")) return TRUE;

	return FALSE;
}

char *COggVorbisPlugin::GetExtensions()
{
	// returns a string containing all supported file extensions + their descriptions
	// format: "description1\0*.ext1;*.ext2;*.ext3\0description2\0*.ext1;*.ext2;*.ext3\0\0"
	// this information will be used in open/add dialogs and with shell integration

	// conclusion: will passed unchanged to Windoz FileDilogs

	return "OggVorbis Files\0*.ogg\0\0";
}

void COggVorbisPlugin::Config()
{
	// opens a configuration dialog
	// the plug-in may store its configuration wherever it desires

	MessageBox(GetActiveWindow(),"Nothing to do","Config...",MB_OK);
}

void COggVorbisPlugin::About()
{
	// opens an about dialog

	MessageBox(GetActiveWindow(),
	"OggVorbis Plug-in by Philips, (c) 2001" "\n"
	"based on Illustrative raw PCM input plug-in by Heikki Ylinen"
#ifdef VORBIS_SDK
	"\n" VORBIS_SDK " for Win32"
#endif /* VORBIS_SDK */
	,"About...",MB_OK);
}


//
//
//
//
//


COggVorbisDecoder::COggVorbisDecoder(char *filename,int audioDataOffset)
{

	DBG("Call COggVorbisDecoder::COggVorbisDecoder", filename);

	_buffer = NULL;
	_beginOfRAWData = audioDataOffset;
	_hFile = INVALID_HANDLE_VALUE;

	if (strstr(filename,"://") != NULL) return;		 // we can't handle URLs

	_hFile = CreateFile(filename,GENERIC_READ,FILE_SHARE_READ | FILE_SHARE_WRITE,NULL,OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL,NULL);

	if (_hFile == INVALID_HANDLE_VALUE) return;		  // error opening file

	// init call backs structure of this instance
	_cb.read_func = &vfcb_i_read;
	_cb.seek_func = &vfcb_i_seek;
	_cb.close_func = &vfcb_i_close;
	_cb.tell_func = &vfcb_i_tell;

	_v_current_section = 0; // !!! not checked - docs have nothing about start value of this

	if (ov_open_callbacks((void *) &_hFile, &_vf, NULL, 0, _cb)) {
		CloseHandle(_hFile);
		_hFile = INVALID_HANDLE_VALUE;
		return;
	};

	DBG("Plugin main init was succesfull!", "DBG");

//	_playingTime = (int)((__int64)(GetFileSize(_hFile,NULL) - _beginOfRAWData) * 1000 / (SAMPLERATE * NUMCHANNELS * (BITRESOLUTION >> 3)));
	_buffer = new char[PCMCHUNKSIZE];

	// seek to the beginning of the actual data

	SetFilePointer(_hFile,_beginOfRAWData,NULL,FILE_BEGIN);
}

COggVorbisDecoder::~COggVorbisDecoder()
{
	// destructor - do destruction, devastation.
	ov_clear(&_vf);
	if (_hFile != INVALID_HANDLE_VALUE) CloseHandle(_hFile);
	if (_buffer != NULL) delete [] _buffer;
}

int COggVorbisDecoder::GetNextAudioChunk(void **buffer)
{
	if (_hFile == INVALID_HANDLE_VALUE) return 0;
		// invalid file, return 0 to indicate end of data

	// DWORD numBytes;
	// ReadFile(_hFile,_buffer,PCMCHUNKSIZE,&numBytes,NULL);

	long bytes_read = OV_HOLE;

        // try to rewind to skip errors/noise in stream
        while (bytes_read == OV_HOLE || bytes_read == OV_EBADLINK) {
	        bytes_read = ov_read(&_vf, _buffer, PCMCHUNKSIZE, 0, 2, 1, &_v_current_section);
	};

	// if we at the get 0 bytes read - going out...  )-:
	if (bytes_read == 0 || bytes_read == OV_HOLE || bytes_read == OV_EBADLINK) {
		return 0;
	};

	// set the buffer-parameter to point to our data
	// the player will not manipulate this data

	*buffer = _buffer;

	// return the data size in bytes
	// the size of decoded data may vary from call to call,
	// but it must not be over 50K  // why not <64k?

	// returning 0 means end of data

	return bytes_read;
}

int COggVorbisDecoder::SetPosition(int newPosition)
{
	// the position is given in milliseconds


	int position = newPosition;

	if (ov_time_seek(&_vf, 0.0+position/1000)) {
		return 0;	// some error
	};
/*
	if (position > _playingTime) position = _playingTime;
	if (position < 0) position = 0;

	// round the seek position appropriately

	position *= SAMPLERATE * NUMCHANNELS * (BITRESOLUTION >> 3);
	position /= 1000;
	position /= NUMCHANNELS * (BITRESOLUTION >> 3);
	position *= NUMCHANNELS * (BITRESOLUTION >> 3);

	SetFilePointer(_hFile,_beginOfRAWData + (int)position,NULL,FILE_BEGIN);

	// calculate the actual seek point in milliseconds and return it

	position *= 1000;
	position /= SAMPLERATE * NUMCHANNELS * (BITRESOLUTION >> 3);
*/
	return position;
}

void COggVorbisDecoder::AdjustEqualizer(int equalizerValues[16])
{
	// the equlizer values are in decibels
	// their respective frequencies in Hz are
	// 50,130,220,300,430,600,820,1160,1680,2370,3320,4690,6680,9430,13300,18800

	// we don't need to do anything here since we set trackInfo->hasEqualizer = FALSE
}

void COggVorbisDecoder::SetEqualizer(BOOL equalize)
{
	// turns the equalizer on/off

	// we don't need to do anything here since we set trackInfo->hasEqualizer = FALSE
}
